//-------------------------------------------------------------------------
/*
Copyright (C) 2005 - Justin Marshall

This file is part of the blood reconstruction project.

WinBlood is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
//-------------------------------------------------------------------------

// ai_Threads.cpp
/***********************************
AI has its own thread system,
because the other thread system
depends on a script to be present.

Eventually we will use this as a thread
that gets spawned from this thread. But
for now its only this with the hardcoded
AI.
************************************/

#include "game.h"

vector<HANDLE> aiActive;

/***************************************
Function: void ai_Shutdown(void)

Info: Shutsdown all AI
***************************************/

void ai_Shutdown(void)
{
	for(int i = 0; i < aiActive.size(); i++)
	{
		TerminateThread(aiActive.at(i),0);
	}
	aiActive.clear();
}

/***************************************
Function DWORD WINAPI aiThreadFunc( LPVOID lpParam ) 

Info: Calls think main, and runs AI
in thread.  Loop end means mob is dead.
***************************************/

DWORD WINAPI aiThreadFunc( LPVOID lpParam ) 
{
	aiConstruct thisAI = *(aiConstruct *)lpParam;
	assert(thisAI.mesprite < thisGame->numofsprites);

	switch(thisAI.aiNumber)
	{
		case MONSTER_ZOMBIE:
			aiZombie *thisZombie  = new aiZombie();
			thisZombie->spritenum =  thisAI.mesprite;
			thisZombie->thinkMain();
		break;
	}

	
	return 0;
}

/**************************************
Function: void ai_SpawnThread(aiConstruct thisAI)

Info: Spawns a thread and calls thinkMain
**************************************/
void ai_SpawnThread(aiConstruct thisAI)
{
	DWORD  dwThreadId;
	HANDLE newThread;
	LPVOID thisData = new LPVOID[sizeof(aiConstruct)];
	memcpy(thisData,&thisAI,sizeof(aiConstruct));
	newThread = CreateThread(NULL,0,aiThreadFunc,thisData,0,&dwThreadId);
	aiActive.push_back(newThread);
}