//-------------------------------------------------------------------------
/*
Copyright (C) 2005 - Justin Marshall

This file is part of the blood reconstruction project.

WinBlood is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

PORTIONS OF CODE FROM DUKE NUKEM 3D SOURCE CODE
*/
//-------------------------------------------------------------------------
// soundManager.cpp

#include "game.h"

soundManager_t::soundManager_t()
{
	printf("....loading sound system\n");
	failedInitSound = FALSE;
	if(!FSOUND_Init(41000, 32, 0))
	{
		printf("WARNING: FSound failed to startup.  Sound will be disabled\n");
		failedInitSound = TRUE;
	}
};

/*
====================================================
int soundManager_t::precache_Sound(char *filename, int hertz, int looping)
====================================================
*/

int soundManager_t::precache_Sound(char *filename, int hertz, int looping, int volume)
{
	FSOUND_SAMPLE *thisSound;
	string		  finalPath;
	unsigned int  playmode;
	soundInfo_t   *worldID   = new soundInfo_t;

	if(failedInitSound == TRUE)
		return -1;
	// see if the sound is allready loaded
	for(int i = 0; i < worldSounds.size(); i++)
	{
	//	if(!strcmp(filename,worldSounds.at(i)->soundName.c_str()))
	//		return i;
	}

	worldID->soundName  = filename;	
	// Construct the file path
	finalPath       = ".\\raw\\";
	finalPath.append(filename);

	playmode = FSOUND_UNSIGNED | FSOUND_8BITS | FSOUND_FORCEMONO | FSOUND_LOADRAW;

	if(looping)
		playmode |= FSOUND_LOOP_NORMAL;

	thisSound = FSOUND_Sample_Load(FSOUND_FREE, finalPath.c_str(), playmode , 0, 0);

//	sndvolume.push_back(volume);
//	thisSound = FSOUND_Stream_Open(finalPath.c_str(), FSOUND_2D | FSOUND_LOADRAW,0,0);

	if(!thisSound)
	{
		printf("SoundManager: Failed to load sound %s\n",finalPath.c_str());
		return -1;
	}

	FSOUND_Sample_SetDefaults(thisSound,hertz,volume,-1,-1);

	worldSounds.push_back(worldID);
	cachedSounds.push_back(thisSound);
	return worldSounds.size()-1;
}

/*
=========================================
void soundManager_t::FlushSystem(void)
=========================================
*/


void soundManager_t::FlushSystem(void)
{
	worldSounds.clear();
	cachedSounds.clear();
	activeSounds.clear();   
}

/*
=========================================================
void soundManager_t::playsnd3d(int spriteid, int sndID)
=========================================================
*/

void soundManager_t::playsnd3d(int spriteid, int sndID)
{
	bloodPlayer_t   *thisPlayer = thisGame->gamePlayers.at(0);
	spritetype      *thisSprite = build_func->Game_GetSprite(spriteid);
	xspritetype     *xsprite    = build_func->GetXSprite(spriteid-1);
	playingSounds_t *thissndptr = new playingSounds_t();
	long            rangeX      = xsprite->data1;
	long            rangeY      = xsprite->data2;
	long            distx,disty;

	if(failedInitSound == TRUE)
		return;
	if(sndID > cachedSounds.size())
	{
		printf("WARNING: Sound %d is out of bounds(%d)\n",sndID,cachedSounds.size());
		return;
	}

	distx = thisPlayer->posx - thisSprite->x;
	disty = thisPlayer->posy - thisSprite->y;

	if(FindDistance2D((long)thisPlayer->posx-thisSprite->x, (long)thisPlayer->posy-thisSprite->y) < 4000)
	{
		if(activeSounds.size() > 0)
		{
			for(int i = 0; i < activeSounds.size(); i++) // see if the sound is playing
			{
				if(activeSounds.at(i)->spritenum == spriteid)
						return;
			}
			// if not add it here.
			thissndptr->channel   = 12 + activeSounds.size();
			thissndptr->soundId   = sndID;
			thissndptr->spritenum = spriteid;

			//FSOUND_SetVolume(thissndptr->channel,sndvolume.at(thissndptr->soundId));
			FSOUND_PlaySound(thissndptr->channel,cachedSounds.at(thissndptr->soundId));
			activeSounds.push_back(thissndptr);
		}
		else
		{
			// First sound...
			if(!FSOUND_IsPlaying(12))
			{
				thissndptr->channel   = 12;
				thissndptr->soundId   = sndID;
				thissndptr->spritenum = spriteid;
				//FSOUND_SetVolume(12,sndvolume.at(thissndptr->soundId));
				FSOUND_PlaySound(12,cachedSounds.at(thissndptr->soundId));
				activeSounds.push_back(thissndptr);
			}
		}
	}
	else
	{
		for(int i = 0; i < activeSounds.size(); i++)
		{
			if(activeSounds.at(i)->spritenum == spriteid)
			{
				// remove this sound no longer needed
				FSOUND_StopSound(activeSounds.at(i)->channel);
				activeSounds.erase(activeSounds.begin() + i);
				break;
			}
		}
	}
}

/*
==========================================
void soundManager_t::StopAllsounds(void)
==========================================
*/

void soundManager_t::StopAllsounds(void)
{
	FMUSIC_StopAllSongs();
	FSOUND_StopSound(FSOUND_ALL);
}

/*
===========================================
void soundManager_t::soundPlay(int sndID)
===========================================
*/

void soundManager_t::soundPlay(int sndID)
{
	if(failedInitSound == TRUE)
		return;
	if(sndID > cachedSounds.size())
	{
		printf("WARNING: Sound %d is out of bounds(%d)\n",sndID,cachedSounds.size());
		return;
	}
	//if(!FSOUND_SetFrequency(FSOUND_ALL,11025))
	//	printf("WARNING: Set frequency failed\n");
	FSOUND_PlaySound(FSOUND_FREE,cachedSounds.at(sndID));
}

/*
====================================
int cacheMusic(char *filename)
====================================
*/

int soundManager_t::cacheMusic(char *filename)
{
	FMUSIC_MODULE *thisSong;
	string		  finalPath;
	soundInfo_t   *worldID   = new soundInfo_t;

	if(failedInitSound == TRUE)
		return -1;
	// see if the sound is allready loaded
	for(int i = 0; i < worldSounds.size(); i++)
	{
		if(!strcmp(filename,worldSounds.at(i)->soundName.c_str()))
			return i;
	}

	worldID->soundName  = filename;

	// Construct the file path
	finalPath       = ".\\mid\\";
	finalPath.append(filename);
	finalPath.append(".mid");

	// DEBUG only remove in non testing builds
	//printf(".....Loading song %s\n",finalPath.c_str());

	thisSong = FMUSIC_LoadSong(finalPath.c_str());

	if(!thisSong)
	{
		thisGame->Error("Failed to load a required music file\n");
	}
	
	// ID's for these two must be the same
	cachedMusic.push_back(thisSong);
	//worldSounds.push_back(worldID);
	return cachedMusic.size()-1;
}

/*
=============================================
void soundManager_t::PlayMusic(int musicID)
=============================================
*/

void soundManager_t::PlayMusic(int musicID)
{
	if(failedInitSound == TRUE)
		return;

	FMUSIC_SetMasterVolume(cachedMusic.at(musicID),140);
	FMUSIC_PlaySong(cachedMusic.at(musicID));
}